#Learning imports
from decision_api.decision_tree_gr import DecisionTreeGR
from planning.baxter import Baxter
import detecting.model as dm
import cv2
#Baxter and ROS imports
import baxter_interface
import rospy
import time

class BaxterAgent(object):
    """
    Agent that manages the baxter robot learning process and deployment
    """
    def __init__(self, config):
        self.config = config

        self.baxter = Baxter();
        self.baxter.both.move(self.baxter.both.neutral)

        self.is_train = config['is_train']

        self.navs = {'left': baxter_interface.Navigator('left'),
            'right': baxter_interface.Navigator('right'),
            'torso_left': baxter_interface.Navigator('torso_left'),
            'torso_right': baxter_interface.Navigator('torso_right')}

        self.ir = dm.load_model(config['ir_path'], config['ir_x'], config['ir_y'])

        if self.is_train:
            self.tree = DecisionTreeGR(features=['color', 'shape'])
        else :
            self.tree = self.load_decision_tree()

    def extract_features(self, img):
        return dm.predict_features(self.ir, img)

    def _train_sort(self, arm_name, msg):
        """ callback used to sort item. The specified arm will move to take a
            snapshot, extract features out of the image, and add those to the learner.
            Then move the item out of the way
        """
        arm = getattr(self.baxter, arm_name)
        if msg:
            arm.move(arm.top)
            rospy.sleep(2)
            img = arm.take_snapshot()
            try:
                f = self.extract_features(img)
                f['direction'] = arm_name.upper()
                print(f)
                self.tree.add_training_sample(f)
                self.baxter.sort(arm_name)
            except:
                print('no features found')
                arm.move(arm.neutral)

    def _train_end(self, msg):
        """ callback used to end training. The learner will then use the samples
            previously taken to train the model
        """
        if msg:
            self.is_train = False
            try:
                self.tree.train()
                self.save_decision_tree()
            except Exception as e:
                print(e)
                print('could not train features')

    def train(self):
        self.drawTrainingUI()
        # callback for left-arm wheel-button
        train_left = lambda m: self._train_sort('left', m)
        train_right = lambda m: self._train_sort('right', m)
        # add callback on button click
        self.navs['left'].button0_changed.connect(train_left)
        self.navs['right'].button0_changed.connect(train_right)
        # add callback on back button click
        self.navs['left'].button1_changed.connect(self._train_end)
        self.navs['right'].button1_changed.connect(self._train_end)
        # blocking call until end button is clicked
        while self.is_train:
            pass
        print('done training')

    def _sort(self, msg):
        """ callback used to sort an item. The left arm will take a snapshot,
            extract features and send them to the learner to choose a side to sort.
            The corresponding arm is then used to sort the item
        """
        if msg:
            self.baxter.left.move(self.baxter.left.top)
            img = self.baxter.left.take_snapshot()
            try:
                features = self.extract_features(img)
                cv2.imwrite('tosort.jpg', img)
                print(features)
                direction = self.tree.classify(features).lower()
                print(direction)
                if direction != 'left':
                    self.baxter.left.move(self.baxter.left.neutral)
                self.baxter.sort(direction)
            except Exception as e:
                print(e)
                self.baxter.both.move(self.baxter.both.neutral)

    def _run_end(self, m):
        if m:
            self.is_run = False

    def run(self):
        self.drawProductionUI()
        self.is_run = True
        self.navs['torso_left'].button0_changed.connect(self._sort)
        self.navs['torso_right'].button0_changed.connect(self._sort)
        self.navs['torso_left'].button1_changed.connect(self._run_end)
        self.navs['torso_right'].button1_changed.connect(self._run_end)
        # blocking call until end button is clicked
        while self.is_run:
            pass
        print('done')


    def load_decision_tree(self):
        return self.tree.load(self.config['tree_dir'])

    def save_decision_tree(self):
        self.tree.save(self.config['tree_dir'])

    def init_move(self):
        self.baxter.both.move(self.baxter.both.neutral)
        self.baxter.right.gripper.close()
        rospy.sleep(3)

    def drawTrainingUI(self):
        print('Garbagebot now training. Please choose a side to sort after putting down an object')
        #xdisplay_image.send_image('path/to/initial/image')

    def drawProductionUI(self):
        print('Garbagebot finished training. Please put down an object and let the robot do the sorting')
