import sys
import cv2
import numpy as np
import tensorflow as tf
import tflearn
from os import walk
from tqdm import tqdm
from imgaug import augmenters as iaa
from uuid import uuid4
from math import sqrt

def isolate_shape(img, rect=None, only_shape=True):
    img = cv2.GaussianBlur(img, (5, 5), 0)

    # plt.imshow(img)
    # plt.show(block=False)

    height, width = img.shape[:2]

    # Create a mask holder
    mask = np.zeros(img.shape[:2], np.uint8)

    # Grab Cut the object
    bgdModel = np.zeros((1, 65), np.float64)
    fgdModel = np.zeros((1, 65), np.float64)

    # "Hard Coding the Rect The object must lie within this rect."
    if not rect:
        rect = (300, 0, 100, height / 2)
    cv2.grabCut(img, mask, rect, bgdModel, fgdModel, 2, cv2.GC_INIT_WITH_RECT)
    # grabCut
    mask = np.where((mask == 2) | (mask == 0), 0, 1).astype('uint8')
    img1 = img * mask[:, :, np.newaxis]

    # Get the background
    background = img - img1

    # Change all pixels in the background that are not black to white
    background[np.where((background > [0, 0, 0]).all(axis=2))] = [
        255, 255, 255
    ]

    # Add the background and the image
    final = background + img1

    if only_shape:
        return background
    return final

def name_to_features(name):
    color = {'RED': 0, 'BLUE': 1, 'GREEN': 2}
    shape = {'CUBE': 0, 'SPHERE': 1, 'TRIANGLE': 2}

    colors = np.zeros(len(color.keys()))
    shapes = np.zeros(len(shape.keys()))
    color = [v for k, v in color.items() if name.startswith(k)][0]
    shape = [v for k, v in shape.items() if name.endswith(k)][0]
    colors[color] = 1
    shapes[shape] = 1
    return np.append(colors, shapes)

def out_to_features(out):
    color = ['RED', 'BLUE', 'GREEN']
    shape = ['CUBE', 'SPHERE', 'TRIANGLE']
    # outputs ranked according to probabilities,
    # first 3 are probabilities for color,
    # next 3 are probabilities for shape
    c_i = np.argmax(out[3:])
    s_i = np.argmax(out[:3])
    return {'color': color[c_i], 'shape': shape[s_i]}

def load(path):
    (_, _, filenames) = walk(path).next()
    data = {}
    for n in filenames:
        name = n.split('.')[0]
        if name not in data:
            data[name] = []
        img = cv2.imread(path + n)
        data[name].append(img)
    return data

def augment(x, y, nb=20):
    st = lambda aug: iaa.Sometimes(0.3, aug)
    seq = iaa.Sequential([
        iaa.Fliplr(0.5),
        st(iaa.Crop(percent=(0, 0.10))),
        st(iaa.Add((-20,20))),
        st(iaa.GaussianBlur(sigma=(0.0, 1.0))),
        st(iaa.Affine(
            scale={"x": (0.8, 1.2), "y": (0.8, 1.2)},
            translate_px={"x": (-20, 20), "y": (-20, 20)},
            rotate=(-20, 20)))
        ], random_order=True)
    aug_data = {}
    x_aug = []
    y_aug = []
    for i in tqdm(range(y.shape[0])):
        xi = [x[i] for n in range(nb)]
        yi = [y[i] for n in range(nb)]
        aug = seq.augment_images(xi)
        x_aug.extend(xi)
        y_aug.extend(yi)
    return np.append(x, np.array(x_aug), axis=0), np.append(y, np.array(y_aug), axis=0)

def preprocess_sample(img, crop_x=(210,430), crop_y=(0, 200)):
    # remove background
    img = isolate_shape(img, only_shape=False)
    # crop to focus on item
    img = img[crop_y[0]:crop_y[1], crop_x[0]:crop_x[1], :]
    # downscale image
    img = cv2.resize(img, (0,0), fx=0.5, fy=0.5)
    # normalise image
    img = img/255.
    return img

def preprocess(data):
    x = []
    y = []
    for k, v in data.items():
        x.extend([preprocess_sample(img) for img in v])
        f = name_to_features(k)
        y.extend([f for i in range(len(v))])
    x = np.array(x)
    y = np.array(y)
    return x, y

def create_model(x_shape=[None, 100, 110, 3], out=9, path='model/model.m'):
    # input is the colored image, this will be spread over 2 nets, one for each feature
    color = tflearn.input_data(shape=x_shape)
    # the shape-learning net will change the img to grayscale
    gray = tflearn.custom_layer(color, tf.image.rgb_to_grayscale)
    final = []
    # make the same net for each feature, using conv nets
    for name, net in {'color': color, 'shape': gray}.items():
        net = tflearn.conv_2d(net, 32, (3, 3), activation='relu')
        net = tflearn.max_pool_2d(net, (2,2))
        net = tflearn.conv_2d(net, 32, (3, 3), activation='relu')
        net = tflearn.max_pool_2d(net, (2,2))
        net = tflearn.conv_2d(net, 64, (3, 3), activation='relu')
        net = tflearn.max_pool_2d(net, (2,2))
        net = tflearn.flatten(net)
        net = tflearn.fully_connected(net, 64, activation='relu')
        net = tflearn.dropout(net, 0.5)
        net = tflearn.fully_connected(net, out, activation='sigmoid')
        net = tflearn.regression(net, optimizer='adam', loss='categorical_crossentropy')
        final.append(net)

    # merge both loss functions into one
    final = tflearn.merge(final, 'concat')
    return tflearn.DNN(final, tensorboard_verbose=0) #, checkpoint_path=path)

def train_net(model, x, y):
    model.fit(x, y, n_epoch=60, shuffle=True, show_metric=True, batch_size=50)

def train_model(model, x, y):
    outs = [y[:,3:], y[:,:3]]
    train_net(model, x, outs)

def predict_features(model, img):
    # cv2.imwrite('tosort.jpg', img)
    img = preprocess_sample(img)
    pred = model.predict([img])[0]
    print(pred)
    return out_to_features(pred)

def load_model(path, x_shape, out):
    model = create_model(x_shape, out)
    model.load(path)
    return model

# this file can be used as main in two ways:
#  1. train the network, needs path to dataset as arg
#  2. use a trained network to predict an image.
#    1st arg: path to image to classify
#    2nd arg: path to model weights
if __name__ == '__main__':
    path = '.'
    pred_path = None
    model_path = None
    if len(sys.argv) > 1:
        path = sys.argv[1]
    if len(sys.argv) > 2:
        pred_path = path
        model_path = sys.argv[2]

    if pred_path is None:
        data = load(path)
        x, y = preprocess(data)
        # x, y = augment(x, y, nb=2)

        s = [None]
        s.extend(x.shape[1:])
        model = create_model(x_shape=s, out=y.shape[1]/2)
        train_model(model, x, y)
        model.save(path + '../checkpoint/model.m')
    else:
        model = load_model(model_path, [None, 100, 110, 3], 3)
        img = cv2.imread(pred_path)
        print(predict_features(model, img))
