import sys
import time
import rospy
import baxter_interface
from baxter_pykdl import baxter_kinematics
import moveit_commander
from baxter_interface import Gripper
import moveit_msgs.msg
from geometry_msgs.msg import PoseStamped, Pose, Point, Quaternion
from sensor_msgs.msg import Image
import cv_bridge as cvb
import cv2

class Arm(moveit_commander.MoveGroupCommander):
    def __init__(self, name, **kwargs):
        """ wrap an interface around the MoveIt and ROS robot api's
            each kwarg is a default position the arm can go to
        """
        super(Arm, self).__init__(name)
        for i,v in kwargs.items():
            setattr(self, i, v)

        name = self.get_name()[:-4] #remove _arm from name)
        self.bridge = cvb.CvBridge()
        self.kin = baxter_kinematics(name)
        self.gripper = Gripper(name)

    def current_pose(self):
        return self.kin.forward_position_kinematics()

    def move(self, pose):
        self.set_start_state_to_current_state()
        self.set_pose_target(pose)
        self.go()

    def _msg_to_jpg(self, msg, name):
        try:
            # Convert your ROS Image message to OpenCV2
            cv2_img = self.bridge.imgmsg_to_cv2(msg, "bgr8")
        except cvb.CvBridgeError, e:
            print(e)
        else:
            # Save your OpenCV2 image as a jpeg
            cv2.imwrite(name + '.jpg', cv2_img)

    def _regular_snapshot(self, msg, name='snap', timeout=1):
        self._msg_to_jpg(msg, name)
        time.sleep(timeout)

    def ready_camera(self):
        arm = self.get_name()[:-4] #remove _arm from name
        #callback = lambda m: self._regular_snapshot(m)
        #subscriber = rospy.Subscriber('cameras/'+arm+'_hand_camera/image', Image, callback)
        msg = rospy.wait_for_message('cameras/'+arm+'_hand_camera/image', Image)
        self._msg_to_jpg(msg, 'snap')
        # no necessary in simulator
        #baxter_interface.CameraController(arm+'_hand_camera').open()

    def take_snapshot(self):
        arm = self.get_name()[:-4] #remove _arm from name
        msg = rospy.wait_for_message('cameras/'+arm+'_hand_camera/image', Image)
        try:
            # Convert your ROS Image message to OpenCV2
            cv2_img = self.bridge.imgmsg_to_cv2(msg, "bgr8")
        except cvb.CvBridgeError, e:
            print(e)
        return cv2_img

class Arms(object):
    def __init__(self, name, *args):
        self._all = moveit_commander.MoveGroupCommander(name)
        self._arms = args
        self.neutral = [arm.neutral for arm in self._arms]

    def current_pose(self):

        return self._all.get_current_pose()

    def set_pose_target(self, *args):
        if len(args) is 1:
            args = args[0]
        for i, pose in enumerate(args):
            self._all.set_pose_target(pose, end_effector_link=self._arms[i].get_end_effector_link())

    def clear_pose_target(self):
        for arm in self._arms:
            self._all.clear_pose_target(arm.get_end_effector_link())

    def plan(self):
        self._all.plan()

    def go(self, wait=True):
        self._all.go(wait=wait)

    def move(self, *args):
        self._all.set_start_state_to_current_state()
        if len(args) is 1:
            args = args[0]
        self.set_pose_target(*args)
        self.go()

class Baxter(object):
    def __init__(self):
        """ make a robot composed of 2 arms, with a number of fixed positions
        """
        self.left = Arm('left_arm', neutral=[0.64623441, 0.84175585, 0.0570087, -0.3827256, 0.92221457, 0.02060252, 0.05115608],
                                    up=[0.74667248, 0.02808032, 0.50554368, -0.00295978, 0.97005472, -0.02409497, 0.24167026],
                                    top=[0.68463346, 0.0016183, 0.06230051, 0.00700717, 0.99995422, -0.00275706, -0.00590466],
                                    pick=[7.54538492e-01, 3.44879303e-04, -6.50381151e-02, 5.63301315e-03, 9.99927904e-01, -9.66346662e-03, -4.36723168e-03],
                                    grasp=[0.75390833, -0.00113117, -0.14, 0.00568284, 0.99992606, -0.00980662, -0.0044056])
        self.right = Arm('right_arm',
                         neutral=[0.649446, -0.83735957, 0.06168316, 0.38299498, 0.92211736, -0.02071508, 0.050846],
                         up=[0.74667248, 0.02808032, 0.50554368, -0.00295978, 0.97005472, -0.02409497, 0.24167026],
                         top=[0.68463346, 0.0016183, 0.06230051, 0.00700717, 0.99995422, -0.00275706, -0.00590466],
                         pick=[7.54538492e-01, 3.44879303e-04, -6.50381151e-02, 5.63301315e-03, 9.99927904e-01, -9.66346662e-03, -4.36723168e-03],
                         grasp=[0.75390833, -0.00113117, -0.14, 0.00568284, 0.99992606, -0.00980662, -0.0044056])

        self.both = Arms('both_arms', self.left, self.right)
        self._commander =  moveit_commander.RobotCommander()

        baxter_interface.RobotEnable().enable()
        rospy.sleep(1)

    def _arm_by_name(self, name):
        return self.arms[self.arm_names.index(name)]

    def to_pose(self, *args):
        pos = ['x', 'y', 'z']
        ori = ['x', 'y', 'z', 'w']
        if len(args) is 1:
            args = args[0]
        p = Point(**dict(zip(pos, args[:3])))
        q = Quaternion(**dict(zip(ori, args[-4:])))
        return Pose(position=p, orientation=q)

    def to_list(self, pose):
        pos = ['x', 'y', 'z']
        ori = ['x', 'y', 'z', 'w']
        return [getattr(pose.position, n) for n in pos], [getattr(pose.orientation, n) for n in ori]

    def get_planning_frame(self):
        return self._commander.get_planning_frame()

    def sort(self, arm):
        arm = getattr(self, arm)
        arm.move(arm.pick)
        arm.gripper.open()
        rospy.sleep(1)
        arm.move(arm.grasp)
        rospy.sleep(1)
        arm.gripper.close()
        rospy.sleep(1)
        arm.move(arm.pick)

        arm.move(arm.neutral)
        arm.gripper.open()
